/*
 * Copyright 2016 - 2024 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <ox/fs/fs.hpp>
#include <ox/model/desctypes.hpp>
#include <ox/std/buffer.hpp>
#include <ox/std/size.hpp>

#include <keel/context.hpp>

#include "input.hpp"

namespace turbine {

class Context;

struct ContextDeleter {
	void operator()(Context *p) noexcept;
};

using ContextUPtr = ox::UPtr<Context, ContextDeleter>;

void shutdown(Context &ctx) noexcept;

keel::Context const&keelCtx(Context const&ctx) noexcept;

keel::Context &keelCtx(Context &ctx) noexcept;

inline ox::FileSystem const*rom(Context const&ctx) noexcept {
	return keelCtx(ctx).rom.get();
}

inline ox::FileSystem *rom(Context &ctx) noexcept {
	return keelCtx(ctx).rom.get();
}

struct WrapBase {
	virtual ~WrapBase() = default;
	virtual WrapBase *copyTo(ox::Span<char> s) noexcept = 0;
	virtual operator bool() const noexcept = 0;
};

template<typename T>
struct Wrap: public WrapBase {
	T *data{};
	Wrap(T *pData) noexcept: data(pData) {
	}
	WrapBase *copyTo(ox::Span<char> s) noexcept override {
		oxAssert(s.size() >= sizeof(Wrap), "too small buffer");
		return new(s.data()) Wrap{data};
	}
	operator bool() const noexcept override {
		return data != nullptr;
	}
};

class WrapPtr {
	private:
		WrapBase *m_wrapPtr{};
		ox::Array<char, sizeof(WrapBase) + sizeof(void*)> m_wrapData;
	public:
		template<typename T>
		inline WrapPtr &operator=(T *ptr) noexcept {
			m_wrapPtr = new(m_wrapData.data()) Wrap(ptr);
			return *this;
		}
		inline WrapPtr &operator=(WrapBase &ptr) noexcept {
			if (ptr) {
				m_wrapPtr = ptr.copyTo(m_wrapData);
			} else {
				m_wrapPtr = nullptr;
			}
			return *this;
		}
		inline operator bool() const noexcept {
			return m_wrapPtr && *m_wrapPtr;
		}
		[[nodiscard]]
		inline WrapBase *getWrapBase() const noexcept {
			return m_wrapPtr;
		}
};

void setApplicationDataRaw(Context &ctx, WrapBase &applicationData) noexcept;

template<typename T>
void setApplicationData(Context &ctx, T *applicationData) noexcept {
	Wrap<T> w(applicationData);
	setApplicationDataRaw(ctx, w);
}

[[nodiscard]]
WrapBase *applicationDataRaw(Context &ctx) noexcept;

template<typename T>
[[nodiscard]]
T *applicationData(Context &ctx) noexcept {
	auto const raw = applicationDataRaw(ctx);
	if (!raw) [[unlikely]] {
		return nullptr;
	}
#ifdef OX_BARE_METAL
	auto const out = static_cast<Wrap<T>*>(raw);
#else
	auto const out = dynamic_cast<Wrap<T>*>(raw);
#endif
	oxAssert(out, "Cast failed - wrong type");
	return out->data;
}

void setKeyEventHandler(Context &ctx, KeyEventHandler h) noexcept;

KeyEventHandler keyEventHandler(Context &ctx) noexcept;

}

